@@ -12,14 +12,17 @@
 #include <X11/Xatom.h>
 #include <X11/X.h>
 
+#include <sys/time.h>
+#include <sys/types.h>
 #include <string.h>
 #include <signal.h>
 #include <stdlib.h>
+#include <unistd.h>
+#include <sys/select.h>
 
 #include "config.h"
 
 #ifdef DEBUG
-#include <unistd.h>
 #if defined(HAVE_BACKTRACE)
 	#include <execinfo.h>
 #elif defined(HAVE_PRINTSTACK)
@@ -34,57 +37,39 @@
 #include "embed.h"
 #include "xembed.h"
 #include "xutils.h"
+#include "wmh.h"
 
 #ifndef NO_NATIVE_KDE
 #include "kde_tray.h"
 #endif
 
 #include "settings.h"
+#include "scrollbars.h"
 #include "tray.h"
 
 struct TrayData tray_data;
-
 static int tray_status_requested = 0;
 
 /****************************
  * Signal handlers, cleanup
  ***************************/
-
 void request_tray_status_on_signal(int sig)
 {
-	psignal(sig, NULL);
 	tray_status_requested = 1;
-	/* This message is not handled, instead it will
-	 * force event loop to spin and eventually reach
-	 * periodic tasks handler, which will read
-	 * tray_status_requestd flag and print out 
-	 * tray status */
-	x11_send_client_msg32(tray_data.async_dpy, 
-			tray_data.tray, 
-			tray_data.tray, 
-			tray_data.xa_tray_opcode, 
-			0, 
-			STALONE_TRAY_STATUS_REQUESTED,
-			0, 0, 0);
-	/* Force event delivery */
-	XSync(tray_data.async_dpy, False);
 }
 
 void cleanup()
 {
 	static int clean = 0;
 	static int cleanup_in_progress = 0;
-
 	if (!clean && cleanup_in_progress) {
-		ERR(("forced to die\n"));
+		LOG_ERROR(("forced to die\n"));
 		abort();
 	}
 	if (clean) return;
 	cleanup_in_progress = 1;
-	DBG(6, ("start\n"));
-
 	if (x11_connection_status()) {
-		DBG(8, ("being nice to the icons\n"));
+		LOG_TRACE(("being nice to the icons\n"));
 		/* Clean the list unembedding icons one by one */
 		icon_list_clean_callback(&embedder_unembed);
 		/* Give away the selection */
@@ -94,46 +79,34 @@ void cleanup()
 		 * process */
 		XSync(tray_data.dpy, False);
 		XCloseDisplay(tray_data.dpy);
-		XCloseDisplay(tray_data.async_dpy);
+		tray_data.dpy = NULL;
 	}
 	cleanup_in_progress = 0;
 	clean = 1;
-	DBG(3, ("done\n"));
 }
 
-void dump_core_on_signal(int sig)
-{
-#ifdef DEBUG
-# ifdef HAVE_BACKTRACE
-#   define BACKTRACE_LEN	15
-	void *array[BACKTRACE_LEN];
-	size_t size;
-# endif
-#endif
-	psignal(sig, NULL);
-#ifdef DEBUG
-	DBG(0, ("-- backtrace --\n"));
-# if defined(HAVE_BACKTRACE)
-	size = backtrace(array, BACKTRACE_LEN);
-	DBG(0, ("%d stack frames obtained\n", size));
-	DBG(0, ("printing out backtrace\n"));
-	backtrace_symbols_fd(array, size, STDERR_FILENO);
-# elif defined(HAVE_PRINTSTACK)
-	printstack(STDERR_FILENO);
-# endif
-#endif
-	abort();
-}
+/**************************************
+ * Helper functions
+ **************************************/
 
-void exit_on_signal(int sig)
+/* Print tray status */
+void dump_tray_status()
 {
-	psignal(sig, NULL);
-	/* This is UGLY and is, probably, to be submitted to
-	 * Daily WTF, but it is the only way I found not to 
-	 * use usleep in main event loop. */
-	DBG(8, ("Sending fake WM_DELETE_WINDOW message\n"));
-	x11_send_client_msg32(tray_data.async_dpy, tray_data.tray, tray_data.tray, tray_data.xa_wm_protocols, tray_data.xa_wm_delete_window, 0, 0, 0, 0);
-	XSync(tray_data.async_dpy, False);
+	int grid_w, grid_h;
+	tray_status_requested = 0;
+	layout_get_size(&grid_w, &grid_h);
+	LOG_INFO(("----------- tray status -----------\n"));
+	LOG_INFO(("active: %s\n", tray_data.is_active ? "yes" : "no"));
+	LOG_INFO(("geometry: %dx%d+%d+%d\n",
+			tray_data.xsh.width, tray_data.xsh.height,
+			tray_data.xsh.x, tray_data.xsh.y));
+	if (tray_data.xembed_data.current)
+		LOG_INFO(("XEMBED focus: 0x%x\n", tray_data.xembed_data.current->wid));
+	else
+		LOG_INFO(("XEMBED focus: none\n"));
+	LOG_INFO(("currently managed icons:\n"));
+	icon_list_forall(&print_icon_data);
+	LOG_INFO(("-----------------------------------\n"));
 }
 
 /**************************************
@@ -144,56 +117,47 @@ void exit_on_signal(int sig)
 void add_icon(Window w, int cmode)
 {
 	struct TrayIcon *ti;
-
 	/* Aviod adding duplicates */
 	if ((ti = icon_list_find(w)) != NULL) {
-		DBG(4, ("ignoring second request to embed 0x%x (requested cmode=%d, current cmode=%d\n",
-					w,
-					cmode,
-					ti->cmode));
+		LOG_TRACE(("ignoring second request to embed 0x%x"
+					"(requested cmode=%d, current cmode=%d\n",
+					w, cmode, ti->cmode));
 		return;
 	}
 	/* Dear Edsger W. Dijkstra, I see you behind my back =( */
-	ti = icon_list_new(w, cmode);
-	if (ti == NULL) return;
-
-#ifdef DEBUG
-	DBG(6, ("starting embedding process for 0x%x with cmode=%d\n", w, cmode));
+	if ((ti = icon_list_new(w, cmode)) == NULL) goto failed0;
+	LOG_TRACE(("starting embedding for icon 0x%x, cmode=%d\n", w, cmode));
 	x11_dump_win_info(tray_data.dpy, w);
-#endif
-
+	/* Start embedding cycle */
 	if (!xembed_check_support(ti)) goto failed1;
-	
 	if (ti->is_xembed_supported)
 		ti->is_visible = xembed_get_mapped_state(ti);
 	else
 		ti->is_visible = True;
-
 	if (ti->is_visible) {
 		if (!embedder_reset_size(ti)) goto failed1;
 		if (!layout_add(ti)) goto failed1;
 	}
-	
 	if (!xembed_embed(ti)) goto failed1;
 	if (!embedder_embed(ti)) goto failed2;
-
 	embedder_update_positions(False);
-	tray_update_window_size();
-
-	DBG(0, ("0x%x: icon added as %s\n", ti->wid, ti->is_visible ? "visible" : "hidden"));
-#ifdef DEBUG
-	print_icon_data(ti);
-#endif
-	return;
-
+	tray_update_window_props();
+	/* Report success */
+	LOG_INFO(("added icon %s (wid 0x%x) as %s\n", 
+			x11_get_window_name(tray_data.dpy, ti->wid, "<unknown>"),
+			ti->wid, 
+			ti->is_visible ? "visible" : "hidden"));
+	goto ok;
 failed2:
 	layout_remove(ti);
 failed1:
-#ifdef DEBUG
-	DBG(0, ("0x%x: embedding failed\n", ti->wid));
-	print_icon_data(ti);	
-#endif
 	icon_list_free(ti);
+failed0:
+	LOG_INFO(("failed to add icon %s (wid 0x%x)\n", 
+			x11_get_window_name(tray_data.dpy, ti->wid, "<unknown>"),
+			ti->wid));
+ok:
+	if (settings.log_level >= LOG_LEVEL_TRACE) dump_tray_status();
 	return;
 }
 
@@ -201,18 +165,24 @@ failed1:
 void remove_icon(Window w)
 {
 	struct TrayIcon *ti;
-	
+	char *name;
+	/* Ignore false alarms */
 	if ((ti = icon_list_find(w)) == NULL) return;
-
+	dump_tray_status();
 	embedder_unembed(ti);
 	xembed_unembed(ti);
 	layout_remove(ti);
 	icon_list_free(ti);
-
-	DBG(0, ("0x%x: icon removed\n", w));
-
-	embedder_update_positions(False);
-	tray_update_window_size();
+	LOG_INFO(("removed icon %s (wid 0x%x)\n",
+				x11_get_window_name(tray_data.dpy, ti->wid, "<unknown>"), 
+				w));
+	/* no need to call embedde_update_positions(), as
+	 * scrollbars_click(SB_WND_MAX) will call it */
+	/* XXX: maybe we need a different name for this
+	 * routine instad of passing cryptinc constant? */
+	scrollbars_click(SB_WND_MAX);
+	tray_update_window_props();
+	dump_tray_status();
 }
 
 /* Track icon visibility state changes */
@@ -220,24 +190,18 @@ void icon_track_visibility_changes(Window w)
 {
 	struct TrayIcon *ti;
 	int mapped;
-
 	/* Ignore false alarms */
 	if ((ti = icon_list_find(w)) == NULL || !ti->is_xembed_supported) return;
-
 	mapped = xembed_get_mapped_state(ti);
-
-	DBG(8, ("xembed_is_mapped(0x%x) = %u\n", w, mapped));
-	DBG(8, ("is_visible = %u\n", ti->is_visible));
+	LOG_TRACE(("xembed_is_mapped(0x%x) = %u\n", w, mapped));
+	LOG_TRACE(("is_visible = %u\n", ti->is_visible));
 #ifdef DEBUG
 	x11_dump_win_info(tray_data.dpy, ti->wid);
 #endif
-
 	/* Nothing has changed */
 	if (mapped == ti->is_visible) return;
-
 	ti->is_visible = mapped;
-	DBG(0, ("0x%x: %s\n", w, mapped ? "showing" : "hiding"));
-
+	LOG_INFO(("%s icon 0x%x\n", mapped ? "showing" : "hiding", w));
 	if (mapped) { /* Icon has become mapped and is listed as hidden. Show this icon. */
 		embedder_reset_size(ti);
 		if (!layout_add(ti)) {
@@ -250,7 +214,7 @@ void icon_track_visibility_changes(Window w)
 		embedder_hide(ti);
 	}
 	embedder_update_positions(False);
-	tray_update_window_size();
+	tray_update_window_props();
 }
 
 /* helper to identify invalid icons */
@@ -266,92 +230,56 @@ void kde_icons_update()
 {
 	unsigned long list_len, i;
 	Window *kde_tray_icons;
-
 	if (tray_data.kde_tray_old_mode || 
 		!x11_get_root_winlist_prop(tray_data.dpy, tray_data.xa_kde_net_system_tray_windows, 
 				(unsigned char **) &kde_tray_icons, &list_len)) 
 	{
 		return;
 	}
-
 	for (i = 0; i < list_len; i++) 
 		/* If the icon is not None and is non old, try to add it
 		 * (if the icon is already there, nothing is gonna happen). */
 		if (kde_tray_icons[i] != None && !kde_tray_is_old_icon(kde_tray_icons[i])) 
 		{
-			DBG(8, ("(possibly unembedded) KDE icon 0x%x\n", kde_tray_icons[i]));
+			LOG_TRACE(("found (possibly unembedded) KDE icon %s (wid 0x%x)\n", 
+						x11_get_window_name(tray_data.dpy, kde_tray_icons[i], "<unknown>"),
+						kde_tray_icons[i]));
 			add_icon(kde_tray_icons[i], CM_KDE);
 		}
-
 	XFree(kde_tray_icons);
 }
 #endif
 
+#define PT_MASK_SB	(1L << 0)
+#define PT_MASK_ALL	PT_MASK_SB
+
 /* Perform several periodic tasks */
-void perform_periodic_tasks()
+void perform_periodic_tasks(int mask)
 {
 	struct TrayIcon *ti;
 	/* 1. Remove all invalid icons */
 	while ((ti = icon_list_forall(&find_invalid_icons)) != NULL) {
-		DBG(4, ("0x%x is invalid. removing\n", ti->wid));
+		LOG_TRACE(("icon 0x%x is invalid; removing\n", ti->wid));
 		remove_icon(ti->wid);
 	}
 	/* 2. Print tray status if asked to */
-	if (tray_status_requested) {
-		unsigned int grid_w, grid_h;
-		tray_status_requested = 0;
-		layout_get_size(&grid_w, &grid_h);
-
-#ifdef DEBUG
-#	define SHOWMSG(msg) DBG(0, msg)
-#else
-#	define SHOWMSG(msg)	print_message_to_stderr msg
-#endif
-
-#ifdef DEBUG
-		trace_mode = 1;
-#endif
-		SHOWMSG(("Someone asked for tray status. Here it comes.\n"));
-		SHOWMSG(("===================================\n"));
-		SHOWMSG(("tray status: %sactive\n", tray_data.is_active ? "" : "not "));
-		SHOWMSG(("grid geometry: %dx%d\n", 
-					grid_w / settings.icon_size, 
-					grid_h / settings.icon_size));
-		SHOWMSG(("tray geometry: %dx%d+%d+%d\n",
-				tray_data.xsh.width, tray_data.xsh.height,
-				tray_data.xsh.x, tray_data.xsh.y));
-#ifdef DEBUG
-		{
-			XWindowAttributes xwa;
-			XGetWindowAttributes(tray_data.dpy, tray_data.tray, &xwa);
-			SHOWMSG(("real tray geometry: %dx%d+%d+%d\n",
-						xwa.width, xwa.height, xwa.x, xwa.y));
-		}
-		if (tray_data.xembed_data.current)
-			SHOWMSG(("icon with focus: 0x%x (pointer %p)\n", tray_data.xembed_data.current->wid, tray_data.xembed_data.current));
-		else
-			SHOWMSG(("no icon is focused\n"));
-#endif
-		dump_icon_list();
-		SHOWMSG(("===================================\n"));
-#ifdef DEBUG
-		trace_mode = 0;
-#endif
-#undef SHOWMSG
-	}
-	{
+	if (tray_status_requested) dump_tray_status();
+	/* 3. KLUDGE to fix window size on (buggy?) WMs */
+	if (settings.kludge_flags & KLUDGE_FIX_WND_SIZE) {
 		/* KLUDGE TODO: resolve */
 		XWindowAttributes xwa;
 		XGetWindowAttributes(tray_data.dpy, tray_data.tray, &xwa);
 		if (!tray_data.is_reparented && 
 				(xwa.width != tray_data.xsh.width || xwa.height != tray_data.xsh.height)) 
 		{
-			DBG(8, ("KLUDGE: fixing window size (current: %dx%d, required: %dx%d)\n",
+			LOG_TRACE(("KLUDGE: fixing tray window size (current: %dx%d, required: %dx%d)\n",
 						xwa.width, xwa.height,
 						tray_data.xsh.width, tray_data.xsh.height));
-			tray_update_window_size();
+			tray_update_window_props();
 		} 
 	}
+	/* 4. run scrollbars periodic tasks */
+	if (mask & PT_MASK_SB) scrollbars_periodic_tasks();
 }
 
 /**********************
@@ -360,20 +288,12 @@ void perform_periodic_tasks()
 
 void expose(XExposeEvent ev)
 {
-	if (ev.window == tray_data.tray && settings.parent_bg) 
+	if (ev.window == tray_data.tray && settings.parent_bg && ev.count == 0) 
 		tray_refresh_window(False);
 }
 
 void visibility_notify(XVisibilityEvent ev)
 {
-#if 0 /* Too much flicker, no real benefit :( */
-	if (ev.state == VisibilityUnobscured && ev.window == tray_data.tray) {
-		static int local_state = True;
-		if (local_state) 
-			tray_refresh_window(False);
-		local_state = !local_state;
-	}
-#endif
 }
 
 void property_notify(XPropertyEvent ev)
@@ -382,7 +302,7 @@ void property_notify(XPropertyEvent ev)
 #if defined(DEBUG) && defined(TRACE_PROPS)
 	char *atom_name;
 	atom_name = XGetAtomName(tray_data.dpy, ev.atom);
-	DBG(6, ("atom = %s\n", atom_name));
+	LOG_TRACE(("atom = %s\n", atom_name));
 	XFree(atom_name);
 #endif
 	/* React on wallpaper change */
@@ -397,10 +317,16 @@ void property_notify(XPropertyEvent ev)
 		if (tray_data.is_active) 
 			kde_icons_update();
 		else 
-			DBG(6, ("not updating KDE icons list: tray is not active\n"));
+			LOG_TRACE(("not updating KDE icons list: tray is not active\n"));
 		kde_tray_update_old_icons(tray_data.dpy);
 	}
 #endif
+	/* React on WM (re)starts */
+	if (ev.atom == XInternAtom(tray_data.dpy, _NET_SUPPORTING_WM_CHECK, False)) {
+		ewmh_list_supported_atoms(tray_data.dpy);
+		tray_set_wm_hints();
+		kde_tray_update_fallback_mode(tray_data.dpy);
+	}
 	/* React on _XEMBED_INFO changes of embedded icons
 	 * (currently used to track icon visibility status) */
 	if (ev.atom == tray_data.xembed_data.xa_xembed_info) {
@@ -421,29 +347,25 @@ void property_notify(XPropertyEvent ev)
 					break;
 				}
 		}
-		DBG(6, ("tray is %sreparented\n", tray_data.is_reparented ? "" : "not "));
+		if (nwindows) XFree(windows);
+		LOG_TRACE(("tray was %sreparented\n", tray_data.is_reparented ? "" : "not "));
 	}
 }
 
 void reparent_notify(XReparentEvent ev)
 {
 	struct TrayIcon *ti;
-
 	ti = icon_list_find(ev.window);
 	if (ti == NULL) return;
-
 	/* Reparenting out of the tray is one of non-destructive
 	 * ways to end XEMBED protocol (see spec) */
 	if (ti->is_embedded && ti->mid_parent != ev.parent) {
-		DBG(3, ("initiating unembedding for 0x%x\n", ti->wid));
+		LOG_TRACE(("will now unembed 0x%x\n", ti->wid));
 #ifdef DEBUG
+		print_icon_data(ti);
 		x11_dump_win_info(tray_data.dpy, ev.parent);
-		dump_icon_list();
 #endif
 		remove_icon(ev.window);
-#ifdef DEBUG
-		dump_icon_list();
-#endif
 	}
 }
 
@@ -454,71 +376,76 @@ void client_message(XClientMessageEvent ev)
 #ifdef DEBUG
 	/* Print neat message(s) about this event to aid debugging */
 	char *msg_type_name;
-
 	msg_type_name = XGetAtomName(tray_data.dpy, ev.message_type);
-
 	if (msg_type_name != NULL) {
-		DBG(4, ("message \"%s\"\n", msg_type_name));
+		LOG_TRACE(("message \"%s\"\n", msg_type_name));
 		XFree(msg_type_name);
 	}
-
 	if (ev.message_type == tray_data.xa_wm_protocols) {
 		msg_type_name = XGetAtomName(tray_data.dpy, ev.data.l[0]);
 		if (msg_type_name != NULL) {
-			DBG(6, ("WM_PROTOCOLS message type: %s\n", msg_type_name));
+			LOG_TRACE(("WM_PROTOCOLS message type: %s\n", msg_type_name));
 			XFree(msg_type_name);
 		}
 	}
 #endif
-
 	/* Graceful exit */
 	if (ev.message_type == tray_data.xa_wm_protocols &&
 		ev.data.l[0] == tray_data.xa_wm_delete_window && 
 		ev.window == tray_data.tray)
 	{
-		DBG(3, ("got WM_DELETE message, will now exit\n"));
+		LOG_TRACE(("got WM_DELETE message, will now exit\n"));
 		exit(0);
 	} 
-
 	/* Handle _NET_SYSTEM_TRAY_* messages */
 	if (ev.message_type == tray_data.xa_tray_opcode && tray_data.is_active) {
-		DBG(6, ("this is the _NET_SYSTEM_TRAY_OPCODE(%lu) message\n", ev.data.l[1]));
-	
+		LOG_TRACE(("this is the _NET_SYSTEM_TRAY_OPCODE(%lu) message\n", ev.data.l[1]));
 		switch (ev.data.l[1]) {
 			/* This is the starting point of NET SYSTEM TRAY protocol */
 			case SYSTEM_TRAY_REQUEST_DOCK:
-				DBG(3, ("dockin' requested by 0x%x, serving in a moment\n", ev.data.l[2]));
+				LOG_TRACE(("dockin' requested by window 0x%x, serving in a moment\n", ev.data.l[2]));
 #ifndef NO_NATIVE_KDE
 				if (kde_tray_check_for_icon(tray_data.dpy, ev.data.l[2])) cmode = CM_KDE;
 				if (kde_tray_is_old_icon(ev.data.l[2])) kde_tray_old_icons_remove(ev.data.l[2]);
 #endif
 				add_icon(ev.data.l[2], cmode);
 				break;
-			/* This is a special case added by this implementation.
-			 * STALONETRAY_TRAY_DOCK_CONFIRMED is sent by stalonetray 
+			/* We ignore these messages, since we do not show
+			 * any baloons anyways */
+			case SYSTEM_TRAY_BEGIN_MESSAGE:
+			case SYSTEM_TRAY_CANCEL_MESSAGE:
+				break;
+			/* Below are special cases added by this implementation */
+			/* STALONETRAY_TRAY_DOCK_CONFIRMED is sent by stalonetray 
 			 * to itself. (see embed.c) */
 			case STALONE_TRAY_DOCK_CONFIRMED:
 				ti = icon_list_find(ev.data.l[2]);
 				if (ti != NULL && !ti->is_embedded) {
 					ti->is_embedded = True;
-					DBG(3, ("0x%x: embedding confirmed\n", ti->wid));
+					LOG_TRACE(("embedding confirmed for icon 0x%x\n", ti->wid));
 #ifdef DEBUG
-					dump_icon_list();
+					dump_tray_status();
 #endif
 				}
-				tray_update_window_size();
+				tray_update_window_props();
+				break;
+			/* Dump tray status on request */
+			case STALONE_TRAY_STATUS_REQUESTED:
+				dump_tray_status();
+				break;
+			/* Find icon and scroll to it if necessary */
+			case STALONE_TRAY_REMOTE_CONTROL:
+				ti = icon_list_find(ev.window);
+				if (ti == NULL) break;
+				scrollbars_scroll_to(ti);
 				break;
-			/* We ignore these messages, since we do not show
-			 * any baloons anyways */
-			case SYSTEM_TRAY_BEGIN_MESSAGE:
-			case SYSTEM_TRAY_CANCEL_MESSAGE:
 			default:
 				break;
 		}
 	}
 #ifdef DEBUG
 	if (ev.message_type == tray_data.xa_tray_opcode && !tray_data.is_active)
-		DBG(6, ("ignoring _NET_SYSTEM_TRAY_OPCODE(%lu) message: tray not active\n", tray_data.is_active));
+		LOG_TRACE(("ignoring _NET_SYSTEM_TRAY_OPCODE(%lu) message because tray is not active\n", tray_data.is_active));
 #endif
 }
 
@@ -544,66 +471,57 @@ void configure_notify(XConfigureEvent ev)
 	struct TrayIcon *ti;
 	struct Point sz;
 	XWindowAttributes xwa;
-
 	if (ev.window == tray_data.tray) {
 		/* Tray window was resized */
 		/* TODO: distinguish between synthetic and real configure notifies */
 		/* TODO: catch rejected configure requests */
-		DBG(8, ("window geometry from event: %ux%u+%d+%d\n", ev.width, ev.height, ev.x, ev.y));
-
+		/* XXX: Geometry stuff is a mess. Geometry
+		 * is specified in slots, but almost always is 
+		 * stored in pixels... */
+		LOG_TRACE(("tray window geometry from event: %ux%u+%d+%d\n", ev.width, ev.height, ev.x, ev.y));
 		/* Sometimes, configure notifies come too late, so we fetch real geometry ourselves */
 		XGetWindowAttributes(tray_data.dpy, tray_data.tray, &xwa);
 		x11_get_window_abs_coords(tray_data.dpy, tray_data.tray, &tray_data.xsh.x, &tray_data.xsh.y);
-		DBG(8, ("real window geometry: %dx%d+%d+%d\n", xwa.width, xwa.height, tray_data.xsh.x, tray_data.xsh.y));
+		LOG_TRACE(("tray window geometry from X11 calls: %dx%d+%d+%d\n", xwa.width, xwa.height, tray_data.xsh.x, tray_data.xsh.y));
 		tray_data.xsh.width = xwa.width;
 		tray_data.xsh.height = xwa.height;
-
 		/* Update icons positions */
-		icon_list_forall(&grid2window);
+		/* XXX: internal API is bad. example below */
+		icon_list_forall(&layout_translate_to_window);
 		embedder_update_positions(True);
-
 		/* Adjust window background if necessary */
 		tray_update_bg(False);
 		tray_refresh_window(True);
-	} else if ((ti = icon_list_find(ev.window)) != NULL) {
-		/* Some icon has resized its window */
-		if (ti->cmode == CM_KDE) {
-			/* KDE icons are not allowed to change their size. Reset their size. */
+		tray_update_window_strut();
+		scrollbars_update();
+	} else if ((ti = icon_list_find(ev.window)) != NULL) { /* Some icon has resized its window */
+		/* KDE icons are not allowed to change their size. Reset icon size. */
+		if (ti->cmode == CM_KDE || settings.kludge_flags & KLUDGE_FORCE_ICONS_SIZE) {
 			embedder_reset_size(ti);
 			return;
 		}
+		if (settings.kludge_flags & KLUDGE_FORCE_ICONS_SIZE) return;
 		/* Get new window size */
 		if (!x11_get_window_size(tray_data.dpy, ti->wid, &sz.x, &sz.y)) {
 			embedder_unembed(ti);
 			return;
 		}
-		DBG(3, ("icon window 0x%x was resized, new size: %ux%u, old size: %ux%u\n", ev.window, 
+		LOG_TRACE(("icon 0x%x was resized, new size: %ux%u, old size: %ux%u\n", ev.window, 
 					sz.x, sz.y, ti->l.wnd_sz.x, ti->l.wnd_sz.y));
 		/* Check if the size has really changed */
 		if (sz.x == ti->l.wnd_sz.x && sz.y == ti->l.wnd_sz.y) return;
 		ti->l.wnd_sz = sz;
 		ti->is_resized = True;
 		/* Do the job */
-#if 0
-		if (layout_handle_icon_resize(ti)) {
-			embedder_refresh(ti);
-#ifdef DEBUG
-			print_icon_data(ti);
-#endif
-			embedder_update_positions(False);
-			tray_update_window_size();
-		}
-#else
 		layout_handle_icon_resize(ti);
 		embedder_refresh(ti);
 #ifdef DEBUG
 		print_icon_data(ti);
 #endif
 		embedder_update_positions(False);
-		tray_update_window_size();
-#endif
+		tray_update_window_props();
 #ifdef DEBUG
-		dump_icon_list();
+		dump_tray_status();
 #endif
 	}
 }
@@ -614,24 +532,23 @@ void selection_clear(XSelectionClearEvent ev)
 	if (ev.selection == tray_data.xa_tray_selection) {
 		/* Is it us who has lost the selection */
 		if (ev.window == tray_data.tray) {
-			DBG(0, ("stalonetray has lost _NET_WM_SYSTRAY selection; deactivating\n"));
+			LOG_INFO(("another tray detected; deactivating\n"));
 			tray_data.is_active = False;
 			tray_data.old_selection_owner = XGetSelectionOwner(tray_data.dpy, tray_data.xa_tray_selection);
 			if (!x11_ok()) {
-				DBG(0, ("could not get new selection owner. re-acquiring selection\n"));
+				LOG_INFO(("could not find proper new tray; reactivating\n"));
 				tray_acquire_selection();
 			};
-				
-			DBG(4, ("the new selection owner 0x%x\n", tray_data.old_selection_owner));
+			LOG_TRACE(("new selection owner is 0x%x\n", tray_data.old_selection_owner));
 			XSelectInput(tray_data.dpy, tray_data.old_selection_owner, StructureNotifyMask);
 			return;
 		} else if (!tray_data.is_active) {
 			/* Someone else has lost selection and tray is not active --- take over the selection */
-			DBG(0, ("re-reacquiring tray selection\n"));
+			LOG_INFO(("another tray exited; reactivating\n"));
 			tray_acquire_selection();
 		} else {
 			/* Just in case */
-			DBG(0, ("WEIRD: tray is active and someone else has lost tray selection :s\n"));
+			LOG_TRACE(("WEIRD: tray is active and someone else has lost tray selection\n"));
 		}
 	}
 }
@@ -641,13 +558,16 @@ void map_notify(XMapEvent ev)
 #ifndef NO_NATIVE_KDE
 	/* Legacy scheme to handle KDE icons */
 	if (tray_data.kde_tray_old_mode) {
-		Window w = kde_tray_find_icon(tray_data.dpy, ev.window);
-		if (w != None) {
-			DBG(4, ("Legacy scheme for KDE icons: detected KDE icon 0x%x. Adding.\n", w));
-			add_icon(w, CM_KDE);
-			/* TODO: remove some properties to trick ion3 so that it no longer thinks that w is a toplevel. 
-			 * Candidates for removal: 
-			 * 	- WM_STATE */
+		struct TrayIcon *ti = icon_list_find_ex(ev.window);
+		if (ti == NULL) {
+			Window w = kde_tray_find_icon(tray_data.dpy, ev.window);
+			if (w != None) {
+				LOG_TRACE(("Legacy scheme for KDE icons: detected KDE icon 0x%x. Adding.\n", w));
+				add_icon(w, CM_KDE);
+				/* TODO: remove some properties to trick ion3 so that it no longer thinks that w is a toplevel. 
+				 * Candidates for removal: 
+				 * 	- WM_STATE */
+			}
 		}
 	}
 #endif
@@ -657,53 +577,37 @@ void unmap_notify(XUnmapEvent ev)
 {
 	struct TrayIcon *ti;
 	ti = icon_list_find(ev.window);
-	if (ti != NULL && !ti->is_invalid && ti->cmode == CM_KDE) {
-		/* KLUDGE! sometimes KDE icons occasionally 
+	if (ti != NULL && !ti->is_invalid) {
+		/* KLUDGE! sometimes icons occasionally 
 		 * unmap their windows, but do _not_ destroy 
 		 * them. We map those windows back */
-		/* XXX: root cause unidentified */
-		DBG(8, ("Unmap KDE icons KLUDGE for 0x%x\n", ti->wid));
+		/* XXX: not root caused */
+		LOG_TRACE(("Unmap icons KLUDGE executed for 0x%x\n", ti->wid));
 		XMapRaised(tray_data.dpy, ti->wid);
 		if (!x11_ok()) ti->is_invalid = True;
 	}
 }
 
+void my_usleep(useconds_t usec)
+{
+	struct timeval timeout;
+	fd_set rfds;
+	FD_ZERO(&rfds);
+	timeout.tv_sec = 0;
+	timeout.tv_usec = usec;
+	select(1, &rfds, NULL, NULL, &timeout);
+}
+
 /*********************************************************/
-int main(int argc, char** argv)
+/* main() for usual operation */
+int tray_main(int argc, char **argv)
 {
 	XEvent		ev;
-	
-	/* Read settings */
-	tray_init();
-	read_settings(argc, argv);
-
-	/* Register cleanup and signal handlers */
-	atexit(cleanup);
-
-	signal(SIGINT,  &exit_on_signal);
-	signal(SIGTERM, &exit_on_signal);
-	signal(SIGHUP,  &exit_on_signal);
-
-	signal(SIGSEGV, &dump_core_on_signal);
-	signal(SIGQUIT, &dump_core_on_signal);
-
-	signal(SIGUSR1, &request_tray_status_on_signal);
-	
-	/* Open display */
-	if ((tray_data.dpy = XOpenDisplay(settings.display_str)) == NULL ||
-		(tray_data.async_dpy = XOpenDisplay(settings.display_str)) == NULL)
-	{
-		DIE(("could not open display\n"));
-	}
-
-	if (settings.xsync)
-		XSynchronize(tray_data.dpy, True);
-
-	x11_trap_errors();
-
-	/* Interpret those settings that need a display */
+	/* Interpret those settings that need an open display */
 	interpret_settings();
-
+#ifdef DEBUG
+	ewmh_list_supported_atoms(tray_data.dpy);
+#endif
 	/* Create and show tray window */
 	tray_create_window(argc, argv);
 	tray_acquire_selection();
@@ -715,65 +619,174 @@ int main(int argc, char** argv)
 #ifndef NO_NATIVE_KDE
 	kde_icons_update();
 #endif
-
 	/* Main event loop */
 	while ("my guitar gently wheeps") {
-		XNextEvent(tray_data.dpy, &ev);
-		xembed_handle_event(ev);
-		switch (ev.type) {
-		case VisibilityNotify:
-			DBG(7, ("VisibilityNotify (0x%x, state=%d)\n", ev.xvisibility.window, ev.xvisibility.state));
-			visibility_notify(ev.xvisibility);
-			break;
-		case Expose:
-			DBG(7, ("Expose (0x%x)\n", ev.xexpose.window));
-			expose(ev.xexpose);
-			break;
-		case PropertyNotify:
-			DBG(7, ("PropertyNotify(0x%x)\n", ev.xproperty.window));
-			property_notify(ev.xproperty);
-			break;
-		case DestroyNotify:
-			DBG(7, ("DestroyNotify(0x%x)\n", ev.xdestroywindow.window));
-			destroy_notify(ev.xdestroywindow);
-			break;
-		case ClientMessage:
-			DBG(5, ("ClientMessage(from 0x%x?)\n", ev.xclient.window));
-			client_message(ev.xclient);
-			break;
-		case ConfigureNotify:
-			DBG(7, ("ConfigureNotify(0x%x)\n", ev.xconfigure.window));
-			configure_notify(ev.xconfigure);
-			break;
-		case MapNotify:
-			DBG(7, ("MapNotify(0x%x)\n", ev.xmap.window));
-			map_notify(ev.xmap);
-			break;
-		case ReparentNotify:
-			DBG(5, ("ReparentNotify(0x%x to 0x%x)\n", ev.xreparent.window, ev.xreparent.parent));
-			reparent_notify(ev.xreparent);
-			break;
-		case SelectionClear:
-			DBG(5, ("SelectionClear (0x%x has lost selection)\n", ev.xselectionclear.window));
-			selection_clear(ev.xselectionclear);
-			break;
-		case SelectionNotify:
-			DBG(5, ("SelectionNotify\n"));
-			break;
-		case SelectionRequest:
-			DBG(5, ("SelectionRequest (from 0x%x to 0x%x)\n", ev.xselectionrequest.requestor, ev.xselectionrequest.owner));
-			break;
-		case UnmapNotify:
-			DBG(7, ("UnmapNotify(0x%x)\n", ev.xunmap.window));
-			unmap_notify(ev.xunmap);
-			break;
-		default:
+		/* This is ugly and extra dependency. But who cares?
+		 * Rationale: we want to block unless absolutely needed.
+		 * This way we ensure that stalonetray does not show up
+		 * in powertop (i.e. does not eat unnecessary power and
+		 * CPU cycles) 
+		 * Drawback: handling of signals is very limited. XNextEvent()
+		 * does not if signal occurs. This means that graceful
+		 * exit on e.g. Ctrl-C cannot be implemented without hacks. */
+		while (XPending(tray_data.dpy) || tray_data.scrollbars_data.scrollbar_down == -1) {
+			XNextEvent(tray_data.dpy, &ev);
+			if (tray_data.terminated) goto bailout;
+			xembed_handle_event(ev);
+			scrollbars_handle_event(ev);
+			switch (ev.type) {
+			case VisibilityNotify:
+				LOG_TRACE(("VisibilityNotify (0x%x, state=%d)\n", ev.xvisibility.window, ev.xvisibility.state));
+				visibility_notify(ev.xvisibility);
+				break;
+			case Expose:
+				LOG_TRACE(("Expose (0x%x)\n", ev.xexpose.window));
+				expose(ev.xexpose);
+				break;
+			case PropertyNotify:
+				LOG_TRACE(("PropertyNotify(0x%x)\n", ev.xproperty.window));
+				property_notify(ev.xproperty);
+				break;
+			case DestroyNotify:
+				LOG_TRACE(("DestroyNotify(0x%x)\n", ev.xdestroywindow.window));
+				destroy_notify(ev.xdestroywindow);
+				break;
+			case ClientMessage:
+				LOG_TRACE(("ClientMessage(from 0x%x?)\n", ev.xclient.window));
+				client_message(ev.xclient);
+				break;
+			case ConfigureNotify:
+				LOG_TRACE(("ConfigureNotify(0x%x)\n", ev.xconfigure.window));
+				configure_notify(ev.xconfigure);
+				break;
+			case MapNotify:
+				LOG_TRACE(("MapNotify(0x%x)\n", ev.xmap.window));
+				map_notify(ev.xmap);
+				break;
+			case ReparentNotify:
+				LOG_TRACE(("ReparentNotify(0x%x to 0x%x)\n", ev.xreparent.window, ev.xreparent.parent));
+				reparent_notify(ev.xreparent);
+				break;
+			case SelectionClear:
+				LOG_TRACE(("SelectionClear (0x%x has lost selection)\n", ev.xselectionclear.window));
+				selection_clear(ev.xselectionclear);
+				break;
+			case SelectionNotify:
+				LOG_TRACE(("SelectionNotify\n"));
+				break;
+			case SelectionRequest:
+				LOG_TRACE(("SelectionRequest (from 0x%x to 0x%x)\n", ev.xselectionrequest.requestor, ev.xselectionrequest.owner));
+				break;
+			case UnmapNotify:
+				LOG_TRACE(("UnmapNotify(0x%x)\n", ev.xunmap.window));
+				unmap_notify(ev.xunmap);
+				break;
+			default:
 #if defined(DEBUG) && defined(TRACE_EVENTS)
-			DBG(8, ("Unhandled event: %s, serial: %d, window: 0x%x\n", x11_event_names[ev.type], ev.xany.serial, ev.xany.window));
+				LOG_TRACE(("Unhandled event: %s, serial: %d, window: 0x%x\n", x11_event_names[ev.type], ev.xany.serial, ev.xany.window));
 #endif
-			break;
+				break;
+			}
+			/* Perform all periodic tasks but for scrollbars */
+			perform_periodic_tasks(PT_MASK_ALL & (~PT_MASK_SB));
 		}
-		perform_periodic_tasks();
+		perform_periodic_tasks(PT_MASK_ALL);
+		my_usleep(500000L);
 	}
+bailout:
+	LOG_TRACE(("Clean exit\n"));
 	return 0;
 }
+
+/* main() for controlling stalonetray remotely */
+int remote_main(int argc, char **argv)
+{
+	Window tray, icon = None;
+	int rc;
+	int x, y, depth = 0, idummy, i;
+	Window win, root;
+	unsigned int udummy, w, h;
+	tray_init_selection_atoms();
+	tray_create_phony_window();
+	LOG_TRACE(("name=\"%s\" btn=%d cnt=%d x=%d y=%d\n", 
+				settings.remote_click_name,
+				settings.remote_click_btn,
+				settings.remote_click_cnt,
+				settings.remote_click_pos.x,
+				settings.remote_click_pos.y));
+	tray = XGetSelectionOwner(tray_data.dpy, tray_data.xa_tray_selection);
+	if (tray == None) return 255;
+	/* 1. find window matching requested name */
+	icon = x11_find_subwindow_by_name(tray_data.dpy, tray, settings.remote_click_name);
+	if (icon == None) return 127;
+	/* 2. form a message to tray requesting it to show the icon */
+	rc = x11_send_client_msg32(tray_data.dpy, /* display */
+			tray, /* destination */
+			icon, /* window */ 
+			tray_data.xa_tray_opcode, /* atom */
+			0, /* data0 */
+			STALONE_TRAY_REMOTE_CONTROL, /* data1 */
+			0, /* data2 */
+			0, /* data3 */
+			0 /* data4 */
+			);
+	if (!rc) return 63;
+	/* 3. Execute the click */
+	/* 3.1. Sort out click position */
+	XGetGeometry(tray_data.dpy, icon, &root, 
+			&idummy, &idummy, 
+			&w, &h, &udummy, &udummy);
+	x = (settings.remote_click_pos.x == REMOTE_CLICK_POS_DEFAULT) ? w / 2 : settings.remote_click_pos.x;
+	y = (settings.remote_click_pos.y == REMOTE_CLICK_POS_DEFAULT) ? h / 2 : settings.remote_click_pos.y;
+	/* 3.2. Find subwindow to execute click on */
+	win = x11_find_subwindow_at(tray_data.dpy, icon, x, y, depth);
+	/* 3.3. Send mouse click(s) to target */
+	LOG_TRACE(("wid=0x%x btn=%d x=%d y=%d\n", 
+				settings.remote_click_name,
+				win,
+				settings.remote_click_btn,
+				settings.remote_click_pos.x,
+				settings.remote_click_pos.y));
+#define SEND_BTN_EVENT(press, time) do { \
+	x11_send_button(tray_data.dpy, /* dispslay */ \
+			press, /* event type */ \
+			win, /* target window */ \
+			root, /* root window */ \
+			time, /* time */ \
+			settings.remote_click_btn, /* button */ \
+			Button1Mask << (settings.remote_click_btn - 1), /* state mask */ \
+			x, /* x coord (relative) */ \
+			y); /* y coord (relative) */ \
+} while (0)
+	for (i = 0; i < settings.remote_click_cnt; i++) {
+		SEND_BTN_EVENT(1, x11_get_server_timestamp(tray_data.dpy, tray_data.tray));
+		SEND_BTN_EVENT(0, x11_get_server_timestamp(tray_data.dpy, tray_data.tray));
+	}
+#undef SEND_BTN_EVENT
+	return 0;
+}
+
+/* main() */
+int main(int argc, char** argv)
+{
+	/* Read settings */
+	tray_init();
+	read_settings(argc, argv);
+	/* Register cleanup and signal handlers */
+	atexit(cleanup);
+	signal(SIGUSR1, &request_tray_status_on_signal);
+	/* Open display */
+	if ((tray_data.dpy = XOpenDisplay(settings.display_str)) == NULL) 
+		DIE(("could not open display\n"));
+	else 
+		LOG_TRACE(("Opened dpy %p\n", tray_data.dpy));
+	if (settings.xsync)
+		XSynchronize(tray_data.dpy, True);
+	x11_trap_errors();
+	/* Execute proper main() function */
+	if (settings.remote_click_name != NULL) 
+		return remote_main(argc, argv);
+	else
+		return tray_main(argc, argv);
+}
+
