@@ -46,14 +46,14 @@ int x11_error_handler(Display *dpy, XErrorEvent *err)
 	XGetErrorText(dpy, trapped_x11_error_code, msg, sizeof(msg)-1);
 	snprintf(req_num_str, 32, "%d", err->request_code);
 	XGetErrorDatabaseText(dpy, "XRequest", req_num_str, "Unknown", req_str, PATH_MAX);
-	DBG(6, ("X11 error: %s (request: %s, resource 0x%x)\n", msg, req_str, err->request_code, err->minor_code, err->resourceid));
+	LOG_ERROR(("X11 error: %s (request: %s, resource 0x%x)\n", msg, req_str, err->request_code, err->minor_code, err->resourceid));
 	return 0;
 }
 
 int x11_ok_helper(const char *file, const int line, const char *func)
 {
 	if (trapped_x11_error_code) {
-		DBG(6, ("X11 error %d detected at %s:%d:%s\n", 
+		LOG_ERROR(("X11 error %d detected at %s:%d:%s\n", 
 					trapped_x11_error_code,
 					file,
 					line,
@@ -87,9 +87,11 @@ static Atom timestamp_atom = None;
 
 Bool x11_wait_for_timestamp(Display *dpy, XEvent *xevent, XPointer data)
 {
-	return (xevent->type == PropertyNotify &&
+	return ((xevent->type == PropertyNotify &&
 	    xevent->xproperty.window == *((Window *)data) &&
-		xevent->xproperty.atom == timestamp_atom);
+		xevent->xproperty.atom == timestamp_atom) ||
+	   (xevent->type == DestroyNotify &&
+		xevent->xdestroywindow.window == *((Window *)data)));
 }
 
 Time x11_get_server_timestamp(Display *dpy, Window wnd)
@@ -105,14 +107,15 @@ Time x11_get_server_timestamp(Display *dpy, Window wnd)
 	XChangeProperty(dpy, wnd, timestamp_atom, timestamp_atom, 8, PropModeReplace, &c, 1);
 	if (!x11_ok()) return CurrentTime;
 
-	/* Wait for the event */
+	/* Wait for the event
+	 * XXX: this may block... */
 	timestamp_wnd = wnd;
 	XIfEvent(dpy, &xevent, x11_wait_for_timestamp, (XPointer)&timestamp_wnd);
 
 	return x11_ok() ? xevent.xproperty.time : CurrentTime;
 }
 
-int x11_get_win_prop32(Display *dpy, Window dst, Atom atom, Atom type, unsigned char **data, unsigned long *len)
+int x11_get_window_prop32(Display *dpy, Window dst, Atom atom, Atom type, unsigned char **data, unsigned long *len)
 {
 	Atom act_type;
 	int act_fmt, rc;
@@ -178,6 +181,42 @@ int x11_send_visibility(Display *dpy, Window dst, long state)
 	return x11_ok() && rc != 0;
 }
 
+int x11_send_button(Display *dpy, 
+		int press, Window dst, Window root, Time time,
+		unsigned int button, unsigned int state, 
+		int x, int y)
+{
+	XEvent ev;
+	int rx, ry, rc, idummy;
+	unsigned int udummy;
+	Window dst_root;
+
+	if (!x11_get_window_abs_coords(dpy, dst, &rx, &ry)) return FAILURE;
+	XGetGeometry(dpy, dst, &dst_root, 
+			&idummy, &idummy, 
+			&udummy, &udummy, &udummy, &udummy);
+	if (!x11_ok()) return FAILURE;
+
+	ev.type = press ? ButtonPress : ButtonRelease;
+	ev.xbutton.display = dpy;
+	ev.xbutton.window = dst;
+	ev.xbutton.subwindow = None;
+	ev.xbutton.root = root;
+	ev.xbutton.time = time;
+	ev.xbutton.x = x;
+	ev.xbutton.y = y;
+	ev.xbutton.x_root = rx + x; 
+	ev.xbutton.y_root = ry + y;
+	ev.xbutton.button = button;
+	ev.xbutton.state = state;
+	ev.xbutton.same_screen = (root == dst_root);
+
+	rc = XSendEvent(dpy, dst, True, 
+			SubstructureNotifyMask | (press ? ButtonPressMask : ButtonReleaseMask),
+			&ev);
+	return rc && x11_ok();
+}
+
 int x11_send_expose(Display *dpy, Window dst, int x, int y, int width, int height)
 {
 	XEvent xe;
@@ -204,33 +243,20 @@ int x11_refresh_window(Display *dpy, Window dst, int width, int height, int expo
 int x11_set_window_size(Display *dpy, Window w, int x, int y)
 {
 	XSizeHints xsh;
-	
 	xsh.flags = PSize;
 	xsh.width = x;
 	xsh.height = y;
-
 	XSetWMNormalHints(dpy, w, &xsh);
 	XResizeWindow(dpy, w, x, y);
-
-	if (!x11_ok()) {
-		DBG(3, ("failed to set 0x%x size to %dx%d\n", w, x, y));
-		return FAILURE;
-	}
-
+	if (!x11_ok()) return FAILURE;
 	return SUCCESS;
 }
 
 int x11_get_window_size(Display *dpy, Window w, int *x, int *y)
 {
 	XWindowAttributes xwa;
-
 	XGetWindowAttributes(dpy, w, &xwa);
-	
-	if (!x11_ok()) {
-		DBG(3, ("failed to get 0x%x attributes\n", w));
-		return FAILURE;
-	}
-
+	if (!x11_ok()) return FAILURE;
 	*x = xwa.width;
 	*y = xwa.height;
 	return SUCCESS;
@@ -243,9 +269,7 @@ int x11_get_window_min_size(Display *dpy, Window w, int *x, int *y)
 	int rc = FAILURE;
 	if (XGetWMNormalHints(dpy, w, &xsh, &flags)) {
 		flags = xsh.flags & flags;
-		DBG(8, ("flags = 0x%x\n", flags));
 		if (flags & PMinSize) {
-			DBG(8, ("min_width = %d, min_height = %d\n", xsh.min_width, xsh.min_height));
 			*x = xsh.min_width;
 			*y = xsh.min_height;
 			rc = SUCCESS;
@@ -258,12 +282,77 @@ int x11_get_window_abs_coords(Display *dpy, Window dst, int *x, int *y)
 {
 	XWindowAttributes xwa;
 	Window child;
-	x11_ok(); /* XXX: this should go away ? */
+#if 0
+	x11_ok(); /* XXX: this should go away, shouldn't it? */
+#endif
 	XGetWindowAttributes(dpy, dst, &xwa);
 	XTranslateCoordinates(dpy, dst, xwa.root, 0, 0, x, y, &child);
 	return x11_ok();
 }
 
+char *x11_get_window_name(Display *dpy, Window dst, char *def)
+{
+	static char *name;
+	if (name != NULL) XFree(name);
+	if (!XFetchName(dpy, dst, &name)) name = NULL;
+	return (name != NULL ? name : def);
+}
+
+Window x11_find_subwindow_by_name(Display *dpy, Window tgt, char *name)
+{
+	char *tgt_name = NULL;
+	Window ret = None, *children, dummy;
+	int i;
+	unsigned int nchildren;
+	if (XFetchName(dpy, tgt, &tgt_name) && !strcmp(tgt_name, name)) ret = tgt;
+	if (!x11_ok()) return None;
+	if (tgt_name != NULL) XFree(tgt_name);
+	if (ret != None) return ret;
+	XQueryTree(dpy, tgt, &dummy, &dummy, &children, &nchildren);
+	if (!x11_ok()) return None;
+	for (i = 0; i < nchildren; i++) {
+		ret = x11_find_subwindow_by_name(dpy, children[i], name);
+		if (ret != None) break;
+	}
+	if (children != NULL) XFree(children);
+	return ret;
+}
+
+Window x11_find_subwindow_at(Display *dpy, Window top, int x, int y, int depth)
+{
+	int d, c, bx = 0, by = 0;
+	Window dummy, *children;
+	unsigned int nchildren;
+	Window cur = top;
+	for (d = 1; d < depth; d++) {
+		/* Query children of current window and traverse them starting
+		 * from the top in stacking order (i.e. from the end of the list
+		 * returned by XQueryTree */
+		XQueryTree(dpy, cur, &dummy, &dummy, &children, &nchildren);
+		if (!x11_ok()) goto fail;
+		/* Exit the loop if window has no children */
+		if (nchildren == 0) break;
+		for (c = nchildren-1; c >=0; c--) {
+			XWindowAttributes xwa;
+			XGetWindowAttributes(dpy, children[c], &xwa);
+			if (!x11_ok()) goto fail;
+			/* Check if this child contains the (x,y) point */
+			if (xwa.x + bx <= x && x < xwa.x + xwa.width + bx &&
+			    xwa.y + by <= y && y < xwa.y + xwa.height + by)
+			{
+				cur = children[c];
+				bx += xwa.x; by += xwa.y;
+				break;
+			}
+		}
+		if (children != NULL) XFree(children);
+	}
+	return cur;
+fail:
+	if (children != NULL) XFree(children);
+	return None;
+}
+
 void x11_extend_root_event_mask(Display *dpy, long mask)
 {
 	static long old_mask = 0;
@@ -271,6 +360,14 @@ void x11_extend_root_event_mask(Display *dpy, long mask)
 	XSelectInput(dpy, DefaultRootWindow(dpy), old_mask);
 }
 
+int x11_parse_color(Display *dpy, char *str, XColor *color)
+{
+	int rc;
+	rc = XParseColor(dpy, XDefaultColormap(dpy, DefaultScreen(dpy)), str, color);
+	if (rc) XAllocColor(dpy, XDefaultColormap(dpy, DefaultScreen(dpy)), color);
+	return x11_ok() && rc;
+}
+
 #ifdef DEBUG
 const char *x11_event_names[LASTEvent] = {
 	"unknown0",
@@ -313,10 +410,9 @@ const char *x11_event_names[LASTEvent] = {
 void x11_dump_win_info(Display *dpy, Window wid)
 {	
 #if defined(DEBUG) && defined(ENABLE_DUMP_WIN_INFO)
-	if (settings.dbg_level >= 8) {
+	if (settings.log_level >= LOG_LEVEL_TRACE) {
 		char cmd[PATH_MAX];
-		DBG(8, ("Dumping info for 0x%x\n", wid));
-		snprintf(cmd, PATH_MAX, "xwininfo -size -bits -stats -id 0x%x 1>&2\n", (unsigned int) wid);
+		snprintf(cmd, PATH_MAX, "xwininfo -tree -size -bits -stats -id 0x%x 1>&2\n", (unsigned int) wid);
 		system(cmd);
 		snprintf(cmd, PATH_MAX, "xprop -id 0x%x 1>&2\n", (unsigned int) wid);
 		system(cmd);
